import time
import future.moves.urllib.request as urllib
import base64
import json
import math
import os
import datetime
from PIL import Image
import piexif
from piexif import helper
import io
import requests
from requests.auth import HTTPDigestAuth

from collections import deque
from lxml import html as Html
from generic_module import Module
from py_logging.cestel_logging import log, log_exception
import logging
 
class CameraModule(Module):
   
  def who_am_i(self):
    type_location = "/control/"
    request = urllib.urlopen(self.protocol + self.host + type_location)
    html = request.read()
    if "MOBOTIX M10" in html:
      return "M10"
    elif "MOBOTIX M12" in html:
      return "M12"
    elif "MOBOTIX M15" in html:
      return "M15"
    elif "MOBOTIX M16" in html:
      return "M16"
    return "ERR"
                     
  def __init__(self, name, init_type, args_dict, root_path):  
    Module.__init__(self, name, root_path)
    self.init_type = init_type
    log(self.logger, 'Camera type {0}'.format(self.init_type), sev=logging.DEBUG)
    
    self.host = args_dict["host"]
    self.port = int(args_dict["port"]) if "port" in args_dict else 80
    self.ignore_classes = args_dict["ignore_classes"] if "ignore_classes" in args_dict else dict()
    self.fixed_offset = args_dict["fixed_offset"] if "fixed_offset" in args_dict else dict()
    self.cam_focus = args_dict["cam_focus"] if "cam_focus" in args_dict else dict()
    self.lanes = args_dict["lanes"] if "lanes" in args_dict else list()
    self.story = int(args_dict["story"]) if "story" in args_dict else 0
    self.num_events_to_parse = int(args_dict["num_events_to_parse"]) if "num_events_to_parse" in args_dict else 10
    self.max_time_diff = int(args_dict["max_time_diff"]) if "max_time_diff" in args_dict else 3000
    self.max_wait_time = int(args_dict["max_wait_time"]) if "max_wait_time" in args_dict else 10
     
    self.protocol = "http://"
    #self.type_location = "/control/"
    self.definitive_type = self.init_type
    if self.init_type == "mobotix":
      self.definitive_type = self.who_am_i()

    self.uname = args_dict["uname"] if "uname" in args_dict else ""
    self.passw = args_dict["passw"] if "passw" in args_dict else ""
    self.picture_type = args_dict["picture_type"] if "picture_type" in args_dict else "overview" #anpr, adr, overview
    #self.protocol = "http://"
    self.vehicles_dict = dict()
    
  def generate_path(self, dt, mod):
    pth = self.sites_path
    pth = pth + "/" + self.sname + "/ext"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + mod
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.init_type 
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + self.name 
    if not os.path.exists(pth):
      os.mkdir(pth)
    dt_parts = dt.split("-")
    pth = pth + "/" + dt_parts[0] + "-" + dt_parts[1] + "-" + dt_parts[2] + "-hh-mm-ss"
    if not os.path.exists(pth):
      os.mkdir(pth)
    pth = pth + "/" + dt_parts[0] + "-" + dt_parts[1] + "-" + dt_parts[2] + "-" + dt_parts[3] + "-mm-ss"
    if not os.path.exists(pth):
      os.mkdir(pth)
    #print pth
    return pth
    
  def add_vehicle(self, vehicle, recv_module_name):
    #sub_vehicle = vehicle.getchildren()[0]
    if recv_module_name not in self.vehicles_dict:
      self.vehicles_dict[recv_module_name] = deque()
    self.vehicles_dict[recv_module_name].append(vehicle)
  
  def datetime_to_timestamp(self, dtm):
    import datetime
    dt_object = datetime.datetime.strptime(dtm, "%Y-%m-%d-%H-%M-%S-%f")
    parts = dtm.split("-")
    ms = float(parts[len(parts) - 1])
    return time.mktime(dt_object.timetuple()) + ms / 1000.0
  
  def timestamp_to_datetime(self, ts):
    return datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d-%H-%M-%S-%f")
  
  def crawl(self, root, storage):
    for child in root.getchildren():
      self.crawl(child, storage)
    if len(root.getchildren()) == 0:
      storage.append(root.text)
      
  def download_and_save(self, cam_ev_list, dtm, ts, lane, auth):
    if self.sname == "":
      return 1
    mindif = self.max_time_diff + 1
    mini = 0
    mint = ""
    #log(str(cam_ev_list))
    for i in range(len(cam_ev_list)):
      dif = math.fabs(ts - cam_ev_list[i][1])
      if dif < mindif:
        mini = i
        mindif = dif
        img_ts = cam_ev_list[i][1]
        img_ts_str = str(img_ts)
        ms = img_ts_str[img_ts_str.find(".") + 1:len(img_ts_str)]
        while len(ms) < 3:
          ms = ms + "0"
        mint = datetime.datetime.fromtimestamp(img_ts).strftime("%Y-%m-%d-%H-%M-%S") + "-" + ms
    if mindif > self.max_time_diff:
      return 1
    cnt = 0
    for i in range(mini - self.story, mini + self.story + 1):
      if i >= 0 and i < len(cam_ev_list):
        req_string = test = self.protocol + self.host + cam_ev_list[i][2]
        log(self.logger, req_string, sev=logging.DEBUG)
        if not auth:
          res = requests.get(req_string)
        else:
          res = requests.get(req_string, auth=HTTPDigestAuth(self.uname, self.passw))
        log(self.logger, 'Status code for request above: {}.'.format(res.status_code), sev=logging.DEBUG)
        if not res.ok:
          log(self.logger, 'Request failed with error {}. Message: {}'.format(res.status_code, res.text))
          return 1

        result = res.content

        calc_id = i - mini
        str_id = str(abs(calc_id))
        if len(str_id) < 2:
          str_id = "0" + str_id
        if calc_id < 0:
          str_id = "n" + str_id
        elif calc_id >= 0:
          str_id = "p" + str_id
        appendice = str_id
        
        if mint != "":
          user_comment = helper.UserComment.dump("SiWIM_PHOTO_TS:" + mint)
        
        temp_buff = io.BytesIO()
        log(self.logger, 'Result is of type {0}'.format(type(result)), sev=logging.DEBUG)
        temp_buff.write(result)
        temp_buff.seek(0)

        try:
          im = Image.open(temp_buff)
        except:
          log(self.logger, 'Failed to parse image. Response contained: {}'.format(res.text), sev=logging.WARNING)
          return 1

        try:
          exif_dict = piexif.load(im.info["exif"])
        except:
          log(self.logger, 'Failed to find exif.', sev=logging.WARNING)
          exif_dict = {}
          exif_dict["Exif"] = {}
          
        exif_dict["Exif"][piexif.ExifIFD.UserComment] = user_comment
        exif_bytes = piexif.dump(exif_dict)
        
        pth = self.generate_path(dtm, "photo")

        try:
          im.save(pth + "/" + dtm + "_" + lane + "_" + appendice + ".jpg", "jpeg", exif=exif_bytes, quality="keep", optimize=True)
        except:
          log(self.logger, 'Failed to save as image. Saving binary blob: {}/{}_{}.blob'.format(pth, dtm, lane), sev=logging.WARNING)
          with open('{}/{}_{}.blob'.format(pth, dtm, lane), 'bw') as blob:
            blob.write(exif_bytes)
          return 1
      cnt+=1
      
    return 0
  
  def wait_for_future(self, ts):
    now = time.time()
    cnt = 0
    while (now - ts - 3) < 0:
      time.sleep(1)
      now = time.time()
      cnt+=1
      if cnt == self.max_wait_time:
        break
    
  #versions 10 and 12  
  def get_mobotix_1(self, veh):
    self.vehicles_dict
    event_list_loc = "/control/player?eventlist"
    event_info_loc = "/control/eventplayer?get_image_info="
    
    sub_veh = veh.getchildren()[0]
    
    lane = int(sub_veh.find("lane").text)
    
    #get picture only for defined lanes
    if lane not in self.lanes:
      return 1

    dtm = sub_veh.find("ts").text    
    v = float(sub_veh.find("v").text)
    
    cls = int(sub_veh.find("cls").text)
        
    #if class isn't interesting, return NOTHING
    if lane in self.ignore_classes and cls in self.ignore_classes[lane]:
      return 1
    
    #convert event time to timestamp (seconds from epoch)
    ts = self.datetime_to_timestamp(dtm)
    #offsets
    if lane in self.fixed_offset:
      ts = ts + self.fixed_offset[lane] / 1000
      
    if lane in self.cam_focus:
      ts = ts + self.cam_focus[lane] / v
      
    dtm_corr = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d-%H-%M-%S") 
      
    #checking if ts is in the future and waiting accordingly
    self.wait_for_future(ts)
      
    evurl = self.protocol + self.host + event_list_loc
    evurl+="&length=" + str(self.num_events_to_parse)
    evurl+="&s_year=" + dtm_corr.split("-")[0]
    evurl+="&s_month=" + dtm_corr.split("-")[1]
    evurl+="&s_day=" + dtm_corr.split("-")[2]
    evurl+="&s_hour=" + dtm_corr.split("-")[3]
    evurl+="&s_min=" + dtm_corr.split("-")[4]
    evurl+="&s_sec=" + dtm_corr.split("-")[5]
    evurl+="&search=Go"
    
    request = urllib.urlopen(evurl)
    html = request.read()    
    
    html_obj = Html.fromstring(html)
    events = html_obj.xpath('//a[@title="Show event image"]')
    
    #find events
    cam_ev_list = list()
    for event in events:
      storage = list()
      self.crawl(event.getparent().getparent(), storage)      
      nm = storage[0]      
      if len(storage) < 3:
        continue    
      dt = storage[1]
      tm = storage[ 2]

      dttm = dt + " " + tm
      
      pth = event.attrib["href"]

      cam_ev_list.append([nm, dttm, pth]) 
    
    #set accurate timestamp
    for cam_ev in cam_ev_list:
      evurl = self.protocol + self.host + event_info_loc + cam_ev[2]
      #print evurl
      request = urllib.urlopen(evurl)
      html = request.read()
        
      dtm_id = html.find(cam_ev[1])
      acc_dtm = html[dtm_id:dtm_id + len(cam_ev[1]) + 4]
      
      facc_dtm = datetime.datetime.strptime(acc_dtm, "%Y-%m-%d %H:%M:%S.%f").strftime("%Y-%m-%d-%H-%M-%S-%f")
      facc_dtm = facc_dtm[:-3]
      
      cam_ev[1] = self.datetime_to_timestamp(facc_dtm)
      cam_ev[1] = float(cam_ev[1])
      #print cam_ev[1]
      
    self.download_and_save(cam_ev_list, dtm, ts, str(lane), False)
    
    return 0    
  
  #versions 15 and 16
  def get_mobotix_2(self, veh):
    photos_per_sec = 5 
    event_list_loc = "/control/event.jpg?output=alarmlist"
    event_img_loc = "/record/events/"
    sub_veh = veh.getchildren()[0]
    lane = int(sub_veh.find("lane").text)
    
    #get picture only for defined lanes
    if lane not in self.lanes:
      return 1
    dtm = sub_veh.find("ts").text    
    v = float(sub_veh.find("v").text)
    
    cls = int(sub_veh.find("cls").text)

    #if class isn't interesting, return NOTHING
    if lane in self.ignore_classes and cls in self.ignore_classes[lane]:
      return 1
    
    #convert event time to timestamp (seconds from epoch)
    ts = self.datetime_to_timestamp(dtm)
    
    #offsets
    if lane in self.fixed_offset:
      ts = ts + self.fixed_offset[lane] / 1000
      
    if lane in self.cam_focus:
      ts = ts + self.cam_focus[lane] / v

    #checking if ts is in the future and waiting accordingly
    self.wait_for_future(ts)
    
    #getting event list  
    tsofs = self.num_events_to_parse / (2 * photos_per_sec)

    evurl = self.protocol + self.host + event_list_loc
    evurl+="&length=" + str(self.num_events_to_parse)
    evurl+="&searchbytime_start=" + str(ts + tsofs) #TODO; do I need to add bias?
    request = urllib.urlopen(evurl)
    
    #json array of events
    html = request.read()
    json_array = json.loads(html)
    cam_ev_list = list()
    for jso in json_array:
      aiv = jso["alarmimage"]
      tsv = float(jso["timestamp"])
      if len(aiv) != 8:
        continue
      pth = event_img_loc + aiv[0:3] + "/" + aiv[3:6] + "/E00000.jpg"
      cam_ev_list.append((aiv, tsv, pth))

    self.download_and_save(cam_ev_list, dtm, ts, str(lane), False) 
    return 0
  """
  def get_axis(self, veh):
    
    event_list_loc = self.protocol + self.host + ":" + str(self.port) + "/cgi-bin/axis_send.py?" #add number of events & ts
    sub_veh = veh.getchildren()[0]
    
    lane = int(sub_veh.find("lane").text)

    #get picture only for defined lanes
    if lane not in self.lanes:
      return 1

    dtm = sub_veh.find("ts").text    
    v = float(sub_veh.find("v").text)
    
    clss = int(sub_veh.find("cls").text)

    #if class isn't interesting, return NOTHING
    if lane in self.ignore_classes and clss in self.ignore_classes[lane]:
      return 1

    #convert event time to timestamp (seconds from epoch)

    ts = self.datetime_to_timestamp(dtm)

    
    #offsets
    if lane in self.fixed_offset:
      ts = ts + self.fixed_offset[lane]
      
    if lane in self.cam_focus:
      ts = ts + self.cam_focus[lane] / v

    #checking if ts is in the future and waiting accordingly
    self.wait_for_future(ts)

    request = urllib.urlopen(event_list_loc + "c=" + self.name + "&n=" + str(self.num_events_to_parse) + "&ts=" + self.timestamp_to_datetime(ts)[:-3])
    
    #list of picture timestamps
    response = request.read()

    
    lines = response.split()
    cam_ev_list = list()
    
    for line in lines:
      aiv = "rage!"
      tsv = self.datetime_to_timestamp(line[:line.find("<br/>")].strip() + "0")
      pth = ":" + str(self.port) + "/cameras/" + self.name + "/" + line[:line.find("<br/>")].strip() + ".jpg"
      cam_ev_list.append((aiv, tsv, pth)) 

    self.download_and_save(cam_ev_list, dtm, ts, str(lane)) 
    return 0
  """
  
  def get_axis(self, veh):
    #print("lol")
    event_list_loc = self.protocol + self.host + ":" + str(self.port) + "/local/image_retrieve/list.cgi?" #add number of events & ts
    sub_veh = veh.getchildren()[0]
    
    lane = int(sub_veh.find("lane").text)

    #get picture only for defined lanes
    if lane not in self.lanes:
      return 1
    
    dtm = sub_veh.find("ts").text    
    v = float(sub_veh.find("v").text)
    
    clss = int(sub_veh.find("cls").text)

    #if class isn't interesting, return NOTHING
    if lane in self.ignore_classes and clss in self.ignore_classes[lane]:
      return 1

    #convert event time to timestamp (seconds from epoch)
    ts = self.datetime_to_timestamp(dtm)
    
    #offsets
    if lane in self.fixed_offset:
      ts = ts + self.fixed_offset[lane]
      
    if lane in self.cam_focus:
      ts = ts + self.cam_focus[lane] / v

    #checking if ts is in the future and waiting accordingly
    self.wait_for_future(ts)
    
    res = requests.get(event_list_loc + "?n=" + str(self.num_events_to_parse) + "?ts=" + self.timestamp_to_datetime(ts)[:-3], auth=HTTPDigestAuth(self.uname, self.passw))
    #list of picture timestamps
    con = res.content

    lines = str(con)[2:-1].split("<br/>")
    lines = lines[:-1]

    cam_ev_list = list()
    
    for line in lines:
      aiv = "rage!"
      tsv = self.datetime_to_timestamp(line)
      pth = ":" + str(self.port) + "/local/image_retrieve/get.cgi?img=" + line + ".jpeg"
      cam_ev_list.append((aiv, tsv, pth)) 
    self.download_and_save(cam_ev_list, dtm, ts, str(lane), True) 
    return 0
    
  def get_technix(self, veh):
    if self.sname == "":
      return
    sub_veh = veh.getchildren()[0]
    anpr_location = "/ANPRArchive/"
    #adr_location = "/ADRArchive/"
    #print etree.tostring(sub_veh)
    dtm = sub_veh.find("ts").text  
    lane = sub_veh.find("lane").text  
    ip = "\\"
    path = None
    if self.picture_type == "anpr":
      ip = sub_veh.find("anpr_ip").text
      path = self.generate_path(dtm, "anpr")
    elif self.picture_type == "adr":
      ip = sub_veh.find("adr_ip").text
      path = self.generate_path(dtm, "anpr")
    elif self.picture_type == "overview":
      ip = sub_veh.find("trg_ip").text
      path = self.generate_path(dtm, "photo")
    #print ip
    ip = ip.replace("\\", "/")
    #print ip
    #print etree.tostring(sub_veh)
    #print path
    #print self.protocol + self.host + ":" + str(self.port) + anpr_location + ip
    request = urllib.Request(self.protocol + self.host + ":" + str(self.port) + anpr_location + ip)
    base64string = base64.b64encode('%s:%s' % (self.uname, self.passw))
    request.add_header("Authorization", "Basic %s" % base64string)   
    result = urllib.urlopen(request).read()
    fd = open(path + "/" + dtm + "_" + lane + "_p00.jpg", "wb")
    fd.write(result)
    fd.close()
    
  def get_macq(self, veh):
    if self.sname == "":
      return
    
    sub_veh = veh.find("wim")
    anpr_veh = veh.find("anpr")
    if anpr_veh == None:
      return

    image_pth = "http://" + self.host + ":" + str(self.port) + "/meci-L1/discr/images/" + anpr_veh.find("trg_ip").text

    dtm = sub_veh.find("ts").text  
    lane = sub_veh.find("lane").text  
    
    if self.picture_type == "anpr":
      image_pth = "http://" + self.host + ":" + str(self.port) + "/meci-L1/discr/images/" + anpr_veh.find("anpr_ip").text
      #path = self.generate_path(dtm, "photo")

    path = self.generate_path(dtm, "photo")

    res = requests.get(image_pth)
    with open('{0}/{1}_{2}_p00.jpg'.format(path, dtm, lane), 'wb') as f:
      for chunk in res.iter_content(1024):
        f.write(chunk)
    
  def get_cammra(self, veh):  # Not a typo.
    if self.sname == "":
      return
    
    sub_veh = veh.find("wim")
    anpr_veh = veh.find("anpr")
    if anpr_veh == None:
      return

    image_pth = "http://" + self.host + ":" + str(self.port) + anpr_veh.find("trg_ip").text

    dtm = sub_veh.find("ts").text  
    lane = sub_veh.find("lane").text  
    
    #if self.picture_type == "anpr":
      #image_pth = "http://" + self.host + ":" + str(self.port) + "/meci-L1/discr/images/" + anpr_veh.find("anpr_ip").text
      #path = self.generate_path(dtm, "photo")

    path = self.generate_path(dtm, "photo")

    try:
      res = requests.get(image_pth, auth=HTTPDigestAuth('root', 'axis'))
      if res.text != str():  # Only write image if it's not empty
        log(self.logger, 'Saving image of size {}'.format(len(res.text)))
        with open('{0}/{1}_{2}_p00.jpg'.format(path, dtm, lane), 'wb') as f:
          for chunk in res.iter_content(1024):
            f.write(chunk)
      else:
        log(self.logger, 'Could not save image because an empty string was received.', sev=logging.WARNING)
    except:
      log_exception(self.logger, 'General exception:')

  def run(self):
    self.alive = True
    self.end = False
    try:
      while True:
        if self.end == True:
          self.alive = False
          log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
          return
        for key in self.vehicles_dict:
          if len(self.vehicles_dict[key]) > 0:
            sub_veh = self.vehicles_dict[key].popleft()
            if self.definitive_type == "M10" or self.definitive_type == "M12":
              try:
                self.get_mobotix_1(sub_veh)
              except:
                log_exception(self.logger, 'Failed to get mobotix_1:')
            elif self.definitive_type == "M15" or self.definitive_type == "M16":
              try:
                self.get_mobotix_2(sub_veh)
              except:
                log_exception(self.logger, 'Failed to get mobotix_2:')
            elif self.definitive_type == "technix":
              try:
                self.get_technix(sub_veh)
              except:
                log_exception(self.logger, 'Failed to get technix:')
            elif self.definitive_type == "axis":
              try:
                self.get_axis(sub_veh)
              except:
                log_exception(self.logger, 'Failed to get axis:')
            elif self.definitive_type == "macq":
              try:
                self.get_macq(sub_veh)
              except:
                log_exception(self.logger, 'Failed to get macq:')
            elif self.definitive_type == "ffgroup":
              try:
                self.get_cammra(sub_veh)
              except:
                log_exception(self.logger, 'Failed to get ffgroup:')
            else:
              log(self.logger, 'A non-anticipaded camera type.', sev=logging.WARNING)
              self.alive = False
              return
        self.zzzzz(0.5)
    except:
      log_exception(self.logger, 'Fatal error:')
      self.alive = False

    
    #self.who_am_i()
    #print "rap"
      