import time
import socket
import threading

from generic_module import Module
from collections import deque
from lxml import etree
from py_logging.cestel_logging import log, log_exception
import logging

class OutputModule(Module):
  def __init__(self, name, port, backlog, root_path):
    Module.__init__(self, name, root_path)
    self.port = port 
    self.backlog = backlog
    
    self.sname = ""
    self.parser = etree.XMLParser(remove_blank_text=True)
    #client-thread indice, vehicle deque pairs
    self.vehicles_dict = dict()
    self.clients_lock = threading.Lock()
    
    self.sock = None
    
  def set_end(self):
    self.end = True
    #this will initiate shutdown procedure; a bit hacky, but this is how we do things around here
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)  
    s.connect(("127.0.0.1", self.port))
    try:
      s.close()
    except:
      pass
    
  def find_lowest_unused_id(self):
    n = 0
    with self.clients_lock:
      while True:
        if n not in self.vehicles_dict:
          return n
        n+=1
    
  def add_vehicle(self, vehicle, module_name="irrelevant"): 
    #print vehicle.find("wim").find("ts").text
    #print "num of listeners: " + str(len(self.vehicles_dict))
    with self.clients_lock:
      for key in self.vehicles_dict:
        self.vehicles_dict[key].append(vehicle)
        
  def set_site_name(self, sname):
    self.sname = sname
    
  def is_doing_something(self):
    return len(self.vehicles_dict) > 0
        
  def clientthread_xml(self, conn, idx):
    #print "clientthread_start: " + str(id)
    try:
      log(self.logger, 'Starting client thread: ' + str(idx))
      while True:
        if self.end:
          with self.clients_lock:
            del self.vehicles_dict[idx]
          log(self.logger, 'Ending client thread: {}'.format(idx), sev=logging.DEBUG)
          conn.shutdown(socket.SHUT_WR)
          conn.close()
          log(self.logger, 'Client thread {} succesfully ended.'.format(idx))
          return
        if len(self.vehicles_dict[idx]) > 0:
          try:
            with self.clients_lock:
              vehicle = self.vehicles_dict[idx].popleft()
            complete_xml = '<swd version="1"><site><name>' + self.sname + '</name><vehicles>' + etree.tostring(vehicle).decode() + '</vehicles></site></swd>'
            to_send = etree.fromstring(complete_xml, self.parser)
          except:
            log_exception(self.logger, 'Failed to send xml:')
            """
            try:
              self.clients_lock.release()
            except Exception:
              pass
            """
            time.sleep(0.5)
            continue
          #print id
          log(self.logger, 'Sending data on thread {}'.format(idx), logging.DEBUG)
          conn.send(etree.tostring(to_send, pretty_print=False))
        time.sleep(0.5)
    except:
      log(self.logger, 'Closing client thread: ' + str(idx), sev=logging.WARNING)
      """
      try:
        self.clients_lock.release()
      except Exception:
        pass
      """
      #print "Ending thread: " + str(len(self.vehicles_dict)) + "\n"
      with self.clients_lock:
        del self.vehicles_dict[idx]
      conn.close()
      return
    
  def run(self):
    self.alive = True
    self.end = False
    for key, mod in self.downstream_modules_dict.items():
      mod.set_upstream_info(self.name, "data_server_port", self.port)
    try:
      self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
      self.sock.bind(('', self.port))
      self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
      self.sock.listen(self.backlog)
    except:
      log_exception(self.logger, 'Fatal error:')
      self.alive = False
      log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
      return
    try:
      while True:
        conn, addr = self.sock.accept()
        #when there are no child threads, end the "main" thread
        if self.end:
          while True:
            with self.clients_lock:
              if len(self.vehicles_dict) == 0:
                try:
                  conn.shutdown(socket.SHUT_WR)
                  conn.close()
                except:
                  log(self.logger, 'Connection close failed:', sev=logging.DEBUG, exc_info=True)
                self.sock.close()
                self.alive = False
                log(self.logger, 'Thread closed correctly.', sev=logging.DEBUG)
                return
            self.zzzzz(0.5)
        idx = self.find_lowest_unused_id()
        with self.clients_lock:
          self.vehicles_dict[idx] = deque()
        threading.Thread(target=self.clientthread_xml, args=(conn, idx)).start()
    except:
      log_exception(self.logger, 'Fatal error:')
      self.sock.close()
    self.alive = False
